
from .session import Session
from ..utils.OrderedList import OrderedList
from ..utils.Types import PRODUCT,VARIETY,TRANSACTION_TYPE,ORDER_TYPE,VALIDITY,EXCHANGE,INTERVAL
from .websocket_connection.ticker import Ticker
from datetime import datetime,date,time,timedelta


class Trader:
    def __init__(self,user_name,password,totp_key) -> None:
        '''
        This Trader is the way to start the Process, the param cred_dict requires ["user_name","password","totp_key"]

        The "totp_key" can be extracted when creating the two factor authentication in the kite app, click on the QR code and use the same key.
        '''
        self.session = Session(user_name,password,totp_key)
        self.instruments = self.session.instruments
        self.instrument_details = self.session.ins_details_cache
        self.ticker = None
        self.startTicker()
    
    def format_date(self,value):
        if isinstance(value,str):
            return value
        elif isinstance(value,datetime):
            return str(value.date())
        elif isinstance(value,date):
            return str(value)
        else:
            raise Exception('expected value types are ["str", "datetime", "date"]\nthe type of value provided is {}'.format(type(value)))
    
    def historical_data(self,instrument_token,start_date,end_date,interval='5minute'):
        url = 'https://kite.zerodha.com/oms/instruments/historical/{}/{}'.format(instrument_token,interval)
        params = {
            'user_id': self.session.user_name,
            'oi': '1',
            'from': self.format_date(start_date),
            'to': self.format_date(end_date),
        }
        response = self.session.request(url=url,method=Session.TYPE_GET,params=params)
        if response['status'] == 'success':
            return OrderedList(response['data']['candles'])
        else:
            raise Exception(response['message'])
    
    def get_orders(self):
        url = 'https://kite.zerodha.com/oms/orders'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
    
    def get_positions(self):
        url = 'https://kite.zerodha.com/oms/portfolio/positions'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
        
    
    def get_holdings(self):
        url = 'https://kite.zerodha.com/oms/portfolio/holdings'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
    
    def get_margins(self):
        url = 'https://kite.zerodha.com/oms/user/margins'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
    
    def get_profile(self):
        url = 'https://kite.zerodha.com/oms/user/profile/full'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
    
    def check_app_sessions(self):
        url = 'https://kite.zerodha.com/api/user/app_sessions'
        response = self.session.request(url=url,method=Session.TYPE_GET)
        if response['status'] == 'success':
            return response['data']
        else:
            raise Exception(response['message'])
    
    def place_order(self,symbol,exchange,transaction_type,quantity,
                    variety=VARIETY.REGULAR,
                    product=PRODUCT.CNC,
                    order_type=ORDER_TYPE.MARKET,
                    validity=VALIDITY.DAY,
                    price='0',
                    disclosed_quantity='0',
                    trigger_price='0',
                    squareoff='0',
                    stoploss='0',
                    trailing_stoploss='0'):
        payload = {
            'variety': variety,
            'exchange': exchange,
            'tradingsymbol': symbol,
            'transaction_type': transaction_type,
            'order_type': order_type,
            'quantity': quantity,
            'price': price,
            'product': product,
            'validity': validity,
            'disclosed_quantity': disclosed_quantity,
            'trigger_price': trigger_price,
            'squareoff': squareoff,
            'stoploss': stoploss,
            'trailing_stoploss': trailing_stoploss,
            'user_id': self.session.user_name
        }
        url = 'https://kite.zerodha.com/oms/orders/regular'
        response = self.session.request(url=url,method=Session.TYPE_POST,data=payload)
        if response['status'] == 'success':
            return response['data']
        elif response['status'] == 'error':
            raise Exception(response['message'])

    def startTicker(self):
        self.ticker = Ticker(session=self.session)
    
    def get_bnf_expiries(self):
        return list(self.instruments['derivatives']['BANKNIFTY']['derivatives'].keys())

    def get_expiries(self,derivative_name):
        return list(self.instruments['derivatives'][derivative_name]['derivatives'].keys())
    
    def get_derivatives_list(self):
        return list(self.instruments['derivatives'].keys())
    
    def get_strikes(self,derivative_name,expiry):
        return list(self.instruments['derivatives'][derivative_name]['derivatives'][expiry]['options'].keys())

    def get_current_expiries_strikes(self,derivative_name):
        current_expiry = self.get_expiries(derivative_name)
        return self.instruments['derivatives'][derivative_name]['derivatives'][current_expiry[0]]['options']
    
    def get_bnf_current_expiry_strikes(self):
        current_expiry = self.get_bnf_expiries()
        return self.instruments['derivatives']['BANKNIFTY']['derivatives'][current_expiry[0]]['options']
    
    def get_instrument_token_equity(self,symbol):
        symbol = symbol.upper()
        try:
            instrument = self.instruments['underlyer_list']['NSE']['NSE']['EQ'][symbol]
        except KeyError:
            raise Exception(f'{symbol} not found in the list of instruments list, please check the symbol correctly again')
        return {'tradingsymbol': instrument['tradingsymbol'],'instrument_token': instrument['instrument_token']}

    def get_instrument_token_option(self,symbol,expiry,strike,strike_type,no_check=False):
        def check(strike):
            if type(strike) == int:
                strike = '{}.0'.format(strike)
            elif type(strike) == float:
                strike = '{}.0'.format(int(strike))
            elif type(strike) == str:
                if strike[-2:] == '.0':
                    strike = strike
                else:
                    strike = '{}.0'.format(int(strike))
            return strike
        
        symbol = symbol.upper()
        strike = check(strike)
        strike_type = strike_type.upper()
        if not no_check:
            try:
                instrument = self.instruments['derivatives'][symbol]
            except KeyError:
                raise Exception(f'{symbol} not found in the list of instruments list, please check the symbol correctly again')
            try:
                instrument = instrument['derivatives'][expiry]
            except KeyError:
                raise Exception(f'{expiry} not found in the list of {symbol} instruments list, please check the symbol correctly again')
            try:
                instrument = instrument['options'][strike]
            except KeyError:
                raise Exception(f'{strike} not found in the expiry list ({expiry}) of {symbol} instruments list, please check the symbol correctly again')
            try:
                instrument = instrument[strike_type]
            except KeyError:
                raise Exception(f'please check strike type')
            return {'tradingsymbol': instrument['tradingsymbol'],'instrument_token': instrument['instrument_token']}
        else:
            ts = self.instruments['derivatives'][symbol]['derivatives'][expiry]['options'][strike][strike_type]['tradingsymbol']
            insToken = self.instruments['derivatives'][symbol]['derivatives'][expiry]['options'][strike][strike_type]['instrument_token']
            return {'tradingsymbol': ts,'instrument_token': insToken}

    def get_instrument_token_index(self,symbol):
        symbol = symbol.upper()
        try:
            try:
                instrument = self.instruments['underlyer_list']['NSE']['NSE-INDICES']['EQ'][symbol]
            except:
                instrument = self.instruments['underlyer_list']['BSE']['BSE-INDICES']['EQ'][symbol]
        except KeyError:
            raise Exception(f'{symbol} not found in the list of instruments list, please check the symbol correctly again')
        return {'tradingsymbol': instrument['tradingsymbol'],'instrument_token': instrument['instrument_token']}
    
    def get_instrument_token_future(self,symbol,expiry):
        symbol = symbol.upper()
        try:
            instrument = self.instruments['derivatives'][symbol]
        except KeyError:
            raise Exception(f'{symbol} not found in the list of instruments list, please check the symbol correctly again')
        try:
            instrument = instrument['derivatives'][expiry]
        except KeyError:
            raise Exception(f'{expiry} not found in the list of {symbol} instruments list, please check the symbol correctly again')
        return {'tradingsymbol': instrument['tradingsymbol'],'instrument_token': instrument['instrument_token']}

    def get_previous_data(self,instrument_token,days=0,interval=INTERVAL.MINUTE_15):
        today = datetime.now()
        start_date = today - timedelta(days=days)
        return self.historical_data(instrument_token=instrument_token,start_date=start_date,end_date=today,interval=interval)
    
    def get_todays_data(self,instrument_token):
        current_day = datetime.now().date()
        return self.historical_data(instrument_token=instrument_token,start_date=current_day,end_date=current_day,interval=INTERVAL.MINUTE_15)
    


import time
import websocket
import json
from datetime import datetime
import threading
from ...utils.parsing import encodeURIComponent,parse_binary
from ...utils.Types import MODE_STRING,KEYS
from typing import Union, List
import os
import certifi

class Ticker:
    def __init__(self,session):
        self.__private_session = session
        self.__private_input_message = None
        self.__private_websocket_connection = None
        self.connect_to_websocket()
        self.last_msg = None
        self.last_msg_time = None
        print('ticker done')

    def connect_to_websocket(self):
        self.params = {
            'api_key': 'kitefront',
            'user_id': self.__private_session.user_name,
            'enctoken': encodeURIComponent(self.__private_session._enc_token),
            'uid': f'{int(time.time() * 1000)}',
            'user-agent': 'kite3-web',
            'version': '3.0.0',
        }
        address = "wss://ws.zerodha.com/"
        url = self.params_to_url(address)
        if os.environ.get('WEBSOCKET_CLIENT_CA_BUNDLE') is None:
            os.environ['WEBSOCKET_CLIENT_CA_BUNDLE'] = certifi.where()
        ws = websocket.WebSocketApp(url, on_message=self.on_message,on_error=self.on_error,on_close=self.on_close)
        ws.on_open = self.on_open
        self.__private_websocket_connection = ws
        thread = threading.Thread(target=ws.run_forever,daemon=True)
        thread.start()
        print(f'connection successful')

        
    
    def params_to_url(self,base_url):
        url_params = '&'.join([f"{key}={value}" for key, value in self.params.items()])
        return f"{base_url}?{url_params}"
    
    def on_message(self,ws, message):
        self.last_msg = parse_binary(message)
        self.last_msg_time = datetime.now()
        #print(parse_binary(message))

    def on_error(self,ws, error):
        print(error)

    def on_close(self,ws):
        print("### closed ###")
    
    def on_open(self,ws):
        m1 = {"a": KEYS.subscribe ,"v":[256265,260105]}
        m2 = {"a": KEYS.mode,"v":[ MODE_STRING.modeLTPC ,[256265,260105]]}
        ws.send(json.dumps(m1))
        ws.send(json.dumps(m2))
    
    def send_message(self,message):
        self.__private_websocket_connection.send(json.dumps(message))
    
    def subscribe(self, tokens: Union[List[int], int],mode=MODE_STRING.modeLTPC):
        token_type = type(tokens)
        if not (token_type == list or token_type == int):
            raise Exception('tokens need to be in a list of integers or just an integer, \nexample 1:- [256265,260105]\nCurrent Type:{}'.format(type(tokens)))
        if token_type == list:
            m1 = {"a": KEYS.subscribe ,"v":tokens}
            m2 = {"a": KEYS.mode,"v":[ mode ,tokens]}
            self.send_message(m1)
            self.send_message(m2)
        elif token_type == int:
            m1 = {"a": KEYS.subscribe ,"v":[tokens]}
            m2 = {"a": KEYS.mode,"v":[ mode ,[tokens]]}
            self.send_message(m1)
            self.send_message(m2)

    def unsubscribe(self, tokens: Union[List[int], int]):
        token_type = type(tokens)
        if not (token_type == list or token_type == int):
            raise Exception('tokens need to be in a list of integers or just an integer, \nexample 1:- [256265,260105]\nCurrent Type:{}'.format(type(tokens)))
        if token_type == list:
            m1 = {"a": KEYS.unsubscribe ,"v":tokens}
            self.send_message(m1)
        elif token_type == int:
            m1 = {"a": KEYS.unsubscribe ,"v":[tokens]}
            self.send_message(m1)


import requests
from time import sleep
import random
import pyotp
import uuid

class Session:
    TYPE_GET = requests.get
    TYPE_POST = requests.post

    def __init__(self,user_name, password, totp_key) -> None:
        self.user_name = user_name
        self._private__password = password
        self.totp_key = totp_key
        self.cookies = None
        self._app_uuid = str(uuid.uuid4())

        # Headers for initial page load (login)
        # Note: Don't set accept-encoding manually - let requests handle it automatically
        self._login_headers = {
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8',
            'accept-language': 'en-GB,en;q=0.5',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36',
        }

        # Headers for API requests (orders, positions, etc.)
        # Note: Don't set accept-encoding manually - let requests handle it automatically
        self.headers = {
            'accept': 'application/json, text/plain, */*',
            'accept-language': 'en-GB,en;q=0.5',
            'content-type': 'application/x-www-form-urlencoded',
            'origin': 'https://kite.zerodha.com',
            'referer': 'https://kite.zerodha.com/',
            'sec-ch-ua': '"Chromium";v="143", "Not A(Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Linux"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'sec-gpc': '1',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36',
            'x-kite-app-uuid': self._app_uuid,
            'x-kite-userid': user_name,
            'x-kite-version': '3.0.0',
        }

        self._base_url = 'https://kite.zerodha.com'
        self._login_url = 'https://kite.zerodha.com/api/login'
        self._totp_url = 'https://kite.zerodha.com/api/twofa'
        self._enc_token = None
        self.public_token = None
        self.start()
        self._get_instruments()

    
    def start(self):
        # Use a session to persist cookies across requests
        self._session = requests.Session()

        # Step 1: Get initial page to collect Cloudflare cookies
        resp = self._session.get(self._base_url, headers=self._login_headers)
        sleep(random.random() + 0.5)

        print('login: authenticating...')
        # Step 2: Login with username/password
        login_data = {'user_id': self.user_name, 'password': self._private__password}
        resp = self._session.post(self._login_url, headers=self._login_headers, data=login_data)

        if resp.status_code == 200:
            request_id = resp.json()['data']['request_id']
        else:
            try:
                obj = resp.json()
            except:
                raise Exception(f'error status {resp.status_code}: issue in the authentication code')
            raise Exception('{}: {}'.format(obj['status'], obj['message']))

        sleep(random.random() + 0.5)
        print('auto TOTP: verifying...')

        # Step 3: 2FA with TOTP
        totp_data = {
            'user_id': self.user_name,
            'request_id': request_id,
            'twofa_value': pyotp.TOTP(self.totp_key).now(),
            'twofa_type': 'totp',
            'skip_session': ''
        }
        response = self._session.post(self._totp_url, headers=self._login_headers, data=totp_data)

        # Extract tokens from cookies
        self.cookies = self._session.cookies
        self._enc_token = self.cookies.get("enctoken")
        self.public_token = self.cookies.get("public_token")

        # Set authorization header for API requests
        self.headers['authorization'] = 'enctoken {}'.format(self._enc_token)

        # Debug: print cookies
        print(f'Cookies obtained: {list(self.cookies.keys())}')

        done = True
        try:
            success = response.json()['status']
        except:
            done = False

        if done:
            print('Authentication: {}'.format(success))
        else:
            print('Authentication: failed')
    
    def request(self, url, method=TYPE_GET, cookies=True, headers=True, data=None, params=None):
        # Use the session to maintain cookies (including Cloudflare cookies)
        if method == requests.get:
            response = self._session.get(
                url,
                headers=self.headers if headers else None,
                params=params
            )
        elif method == requests.post:
            response = self._session.post(
                url,
                headers=self.headers if headers else None,
                data=data,
                params=params
            )
        else:
            response = method(
                url,
                cookies=self.cookies if cookies else None,
                headers=self.headers if headers else None,
                data=data,
                params=params
            )
        return response.json()

    def _get_instruments(self):
        new_headers = {
            'accept': 'application/json, text/plain, */*',
            'accept-language': 'en-GB,en;q=0.9',
            'kite-iframe': 'kite_iframe_user',
            'origin': 'https://insights.sensibull.com',
            'referer': 'https://insights.sensibull.com/',
            'sec-ch-ua': '"Brave";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Linux"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-site',
            'sec-gpc': '1',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
        }
        response = requests.get('https://oxide.sensibull.com/v1/compute/cache/insights/instrument_metacache/1', headers=new_headers)
        next_resp = requests.get('https://oxide.sensibull.com/v1/compute/cache/insights/underlying_instruments', headers=new_headers)
        print('getting instrument tokens')
        if response.status_code == 200:
            resp = response.json()
            del resp['etag']
            self.instruments = resp
            self.ins_details_cache = next_resp.json()['data']
        else:
            raise Exception('Failed to get instruments')
        
        #sleep(2)
        


from collections import namedtuple

PRODUCT = namedtuple('Product', ['NRML', 'MIS','CNC','CO'])(
    NRML='NRML',
    MIS='MIS',
    CNC='CNC',
    CO='CO')


ORDER_TYPE = namedtuple('OrderType', ['MARKET', 'LIMIT', 'SLM', 'SL'])(
    MARKET='MARKET', 
    LIMIT='LIMIT', 
    SLM='SL-M', 
    SL='SL'
)



VARIETY = namedtuple('VarietyType', ['REGULAR', 'CO', 'AMO', 'ICEBERG', 'AUCTION'])(
    REGULAR='regular', 
    CO='co', 
    AMO='amo', 
    ICEBERG='iceberg', 
    AUCTION='auction'
)


TRANSACTION_TYPE = namedtuple('TransactionType', ['BUY', 'SELL'])(
    BUY='BUY', 
    SELL='SELL'
)


VALIDITY = namedtuple('ValidityType', ['DAY', 'IOC', 'TTL'])(
    DAY='DAY', 
    IOC='IOC', 
    TTL='TTL'
)



POSITION_TYPE = namedtuple('PositionType', ['DAY', 'OVERNIGHT'])(
    DAY='day', 
    OVERNIGHT='overnight'
)


EXCHANGE = namedtuple('ExchangeType', ['NSE', 'BSE', 'NFO', 'CDS', 'BFO', 'MCX', 'BCD'])(
    NSE='NSE', 
    BSE='BSE', 
    NFO='NFO', 
    CDS='CDS', 
    BFO='BFO', 
    MCX='MCX', 
    BCD='BCD'
)


MARGIN = namedtuple('MarginType', ['EQUITY', 'COMMODITY'])(
    EQUITY='equity', 
    COMMODITY='commodity'
)


STATUS = namedtuple('StatusType', ['COMPLETE', 'REJECTED', 'CANCELLED'])(
    COMPLETE='COMPLETE', 
    REJECTED='REJECTED', 
    CANCELLED='CANCELLED'
)

SEGMENT = namedtuple('SegmentType', ['segmentNseCM','segmentNseFO','segmentNseCD','segmentBseCM','segmentBseFO','segmentBseCD','segmentMcxFO','segmentMcxSX','segmentNseIndices','segmentAuction','segmentUS','segmentNseCOM'])(
    segmentNseCM = 1,
    segmentNseFO = 2,
    segmentNseCD = 3,
    segmentBseCM = 4,
    segmentBseFO = 5,
    segmentBseCD = 6,
    segmentMcxFO = 7,
    segmentMcxSX = 8,
    segmentNseIndices = 9,
    segmentAuction = 10,
    segmentUS = 11,
    segmentNseCOM = 12
)

MODE_VALUES = namedtuple('modeType', ['modeFull', 'modeQuote', 'modeLTPC','modeLTP'])(
    modeFull = 1,
    modeQuote = 2,
    modeLTPC = 3,
    modeLTP = 4
)

MODE_STRING = namedtuple('modeType', ['modeFull', 'modeQuote', 'modeLTPC','modeLTP'])(
    modeFull = 'full',
    modeQuote = 'quote',
    modeLTPC = 'ltpc',
    modeLTP = 'ltp'
)

KEYS = namedtuple('keyOptions',['subscribe','mode','unsubscribe'])(
    subscribe='subscribe',
    mode='mode',
    unsubscribe='unsubscribe'
)

INTERVAL = namedtuple('intervalOptions',['MINUTE','MINUTE_2','MINUTE_3','MINUTE_5','MINUTE_10','MINUTE_15','MINUTE_30','HOUR','HOUR_2','HOUR_3','DAY'])(
    MINUTE='minute',
    MINUTE_2='2minute',
    MINUTE_3='3minute',
    MINUTE_5='5minute',
    MINUTE_10='10minute',
    MINUTE_15='15minute',
    MINUTE_30='30minute',
    HOUR='60minute',
    HOUR_2='2hour',
    HOUR_3='3hour',
    DAY='day'
)



from datetime import datetime
import numpy as np
import pickle

def load_dict(path):
    with open(path, 'rb') as f:
        data = pickle.load(f)
    return OrderedList(data)

class OrderedList:
    class _ILOC:
        def __init__(self,parent) -> None:
            self.parent = parent
        
        def __getitem__(self, obj):
            if isinstance(obj, int):
                data_head = {key: [value[obj]] for key, value in self.parent._data.items()}
                return OrderedList(data_head)
            elif isinstance(obj, slice):
                data_head = {key: value[obj.start:obj.stop:obj.step] for key, value in self.parent._data.items()}
                return OrderedList(data_head)

    def __init__(self,list_of_values) -> None:
        if type(list_of_values) is dict:
            self._data = list_of_values
        else:
            self._data = self._basic_method(list_of_values=list_of_values)
        self.columns = list(self._data.keys())
        self.iloc = self._ILOC(self)
    
    def __repr__(self) -> str:
        return self.print_dict_as_table(self._data)
    
    def _basic_method(self,list_of_values):
        mod_func = lambda date_string: datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S%z').replace(tzinfo=None)
        nd = {'datetime':[],'open':[],'high':[],'low':[],'close':[],'volume':[],'oi':[]}
        for row in list_of_values:
            nd['datetime'].append(mod_func(row[0]))
            nd['open'].append(row[1])
            nd['high'].append(row[2])
            nd['low'].append(row[3])
            nd['close'].append(row[4])
            nd['volume'].append(row[5])
            nd['oi'].append(row[6])
        return nd
    
    def print_dict_as_table(self,data,trucation=7):
        prints = []
        max_lengths = [max(len(str(value)) for value in col) for col in data.values()]
        header_row = "  |  ".join(header.center(length) for header, length in zip(data.keys(), max_lengths))
        #print(header_row)
        prints.append(header_row)
        separator = "--+--".join("-" * length for length in max_lengths)
        #print(separator)
        prints.append(separator)
        counter = 0
        more_rows = False
        for i in range(len(next(iter(data.values())))):
            row_values = [str(data[key][i]).center(length) for key, length in zip(data.keys(), max_lengths)]
            row = "  |  ".join(row_values)
            #print(row)
            if counter <= trucation:
                prints.append(row)
            else:
                more_rows = True
            counter += 1
        if more_rows:
            prints.append('... upto {} rows'.format(counter))
        return '\n'.join(prints)

    def head(self, val=5):
        data_head = {key: value[:val] for key, value in self._data.items()}
        return OrderedList(data_head)
    
    def tail(self, val=5):
        data_head = {key: value[-val:] for key, value in self._data.items()}
        return OrderedList(data_head)
    
    def __getitem__(self,list_of_columns):
        if type(list_of_columns) is str:
            data_head = {list_of_columns: self._data[list_of_columns]}
        elif type(list_of_columns) is list:
            data_head = {key: self._data[key] for key in list_of_columns}
        return OrderedList(data_head)
    
    @property
    def values(self):
        data_head = [self._data[key] for key in self._data.keys()]
        return np.array(data_head).T
    
    def save(self, path):
        with open(path, 'wb') as f:
            pickle.dump(self._data, f)
    
    def __len__(self):
        return len(self._data['datetime'])